// LseView.cpp : implementation of the CLseView class
//

#include "stdafx.h"
#include "Lse.h"

#include "LseDoc.h"
#include "LseView.h"
#include "LSysDlg.h"
#include "mainfrm.h"

#include <math.h>
#include <fstream.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//#define _DEBUGDRAW

#pragma warning(disable : 4800)

/////////////////////////////////////////////////////////////////////////////
// FIXME: There is way too much information stored in the CLseView class. Most
// of the information (L-System, angle, depth, etc.) should be stored within 
// the document class, but I couldn't be bothered to change it. Live and learn...
//

#define _deg2rad(a) (a * 3.141592653589793238462643) / 180.0

/////////////////////////////////////////////////////////////////////////////
// CLseView

IMPLEMENT_DYNCREATE(CLseView, CView)

BEGIN_MESSAGE_MAP(CLseView, CView)
	//{{AFX_MSG_MAP(CLseView)
	ON_COMMAND(ID_EDIT_LSYSTEM, OnEditLsystem)
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_MOUSEWHEEL()
	ON_COMMAND(ID_VIEW_OPTIONS, OnViewOptions)
	ON_WM_ERASEBKGND()
	ON_COMMAND(ID_EDIT_DECREASEDEPTH, OnEditDecreaseDepth)
	ON_COMMAND(ID_EDIT_INCREASEDEPTH, OnEditIncreaseDepth)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CLseView construction/destruction

CLseView::CLseView()
{
	m_bMouseDragging = false;
	m_bCalculating = FALSE;

	InitRulesArray();

	m_fAngle = 90;
	m_fInitAngle = 0;
	m_fStepSize = 0.65f;
	m_iSegment = 6;
	m_iMaxDepth = 4;
	m_iPenIndex = 0;

	m_crStart = RGB(255,0,0);
	m_crEnd = RGB(0,0,255);
}

CLseView::~CLseView()
{
}

BOOL CLseView::PreCreateWindow(CREATESTRUCT& cs)
{
	// TODO: Modify the Window class or styles here by modifying
	//  the CREATESTRUCT cs

	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CLseView drawing

void CLseView::OnDraw(CDC* pDC)
{
	CLseDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);

	CRect rect;
	GetClientRect(rect);
	
	m_iSP = 0;

	m_iPenIndex = 0;
	m_dcr = GetRValue(m_crStart);
	m_dcg = GetGValue(m_crStart);
	m_dcb = GetBValue(m_crStart);

#ifdef _DEBUGDRAW
	// This code helps debug, since draw straight onto the DC.
	CPen *oldpen, pen(PS_SOLID, 1, m_crLine);
	oldpen = pDC->SelectObject(&pen);

	pDC->MoveTo(m_cStartPoint);
	pDC->FillSolidRect(0,0,rect.Width(), rect.Height(), m_crBackground);

	DrawLSystems(m_pszAxiom, pDC, m_fInitAngle, 0);
#else 
	CDC memdc;
	CBitmap bmp, *oldbmp;
	CPen *oldpen, pen(PS_SOLID, 1, m_crLine);

	memdc.CreateCompatibleDC(GetDC());
	bmp.CreateDiscardableBitmap(pDC, rect.Width(), rect.Height());
	oldbmp = memdc.SelectObject(&bmp);
	oldpen = memdc.SelectObject(&pen);

	memdc.FillSolidRect(0,0,rect.Width(), rect.Height(), m_crBackground);
	memdc.MoveTo(m_cStartPoint);

	DrawLSystems(m_pszAxiom, &memdc, m_fInitAngle, 0);

	pDC->BitBlt(0,0,rect.Width(), rect.Height(), &memdc, 0, 0, SRCCOPY);

	memdc.SelectObject(oldpen);
	memdc.SelectObject(oldbmp);
#endif
}

int CLseView::Round(double x) {
      return (int)floor(x + 0.5);
}

bool CLseView::PushState(float angle, int x, int y) {
	if (m_iSP++ > LSYS_STACK_SIZE) return false;

	_lstate branch;
	branch.angle = angle;
	branch.pos_x = x;
	branch.pos_y = y;

	m_sStack[m_iSP] = branch;

	return true;
}

void CLseView::PopState(float &angle, CPoint &point) {
	if (m_iSP == -1) return;

	_lstate state = m_sStack[m_iSP--];
	angle = state.angle;
	point = CPoint(state.pos_x, state.pos_y);
}

float CLseView::Increment(float init, float inc) {
	init += inc;

	if (init >= 360)
		return init - 360;

	return init;
}

float CLseView::Decrement(float init, float inc) {
	init -= inc;

	if (init < 0) return 360 + init;

	return init;
}

void CLseView::DrawLSystems(CString lstring, CDC *pDC, float angle, int depth)
{
	char at;									// Cache the character
	int number = 0;								// Number support
	int	len = lstring.GetLength();				// Cache the length
	int old;									// Old segment length

	LOGPEN logpen;
	CPen pen, *oldpen;
	CPoint start = pDC->GetCurrentPosition();

	pen.CreatePen(PS_SOLID, 1, RGB((int)m_dcr, (int)m_dcg, (int)m_dcb));
	oldpen = pDC->SelectObject(&pen);

	for (int i=0;i<len;i++) {
		at = lstring.GetAt(i);
		
		if (depth < m_iMaxDepth && (at >= 'A' && at <= 'Z')) {
			old = m_iSegment;
			m_iSegment = int(m_iSegment * m_fStepSize);
			if (m_iSegment % 2 != 0) m_iSegment--;
			DrawLSystems(Rule(at), pDC, angle, depth+1);
			m_iSegment = old;
		} else if (at == 'F' || at == '|') {
			int mx = Round(start.x + (m_iSegment*cos(_deg2rad(angle))));
			int my = Round(start.y + (m_iSegment*sin(_deg2rad(angle))));
			
			start.x = mx;
			start.y = my;

			if (!m_bCalculating) {
				pDC->LineTo(start);

				m_dcr += m_dr;
				m_dcg += m_dg;
				m_dcb += m_db;

				pen.GetLogPen(&logpen);
				logpen.lopnColor = RGB((int)m_dcr, (int)m_dcg, (int)m_dcb);

			} else {
				m_iSegments++;
			}
		} else if (m_bCalculating) {
			continue;
		} else if (at == 'f' || at == 'G') {
			int mx = Round(start.x + (m_iSegment*cos(_deg2rad(angle))));
			int my = Round(start.y + (m_iSegment*sin(_deg2rad(angle))));
			
			start.x = mx;
			start.y = my;

			pDC->MoveTo(start);			
		} else if (at == '+') {
			number = (number) ? number : 1;
			angle = Increment(angle, m_fAngle * number);
			number = 0;
		} else if (at == '-') {
			number = (number) ? number : 1;
			angle = Decrement(angle, m_fAngle * number);
			number = 0;
		} else if (at == '[') {
			number = 0;
			if (!PushState(angle, start.x, start.y)) break;
		} else if (at == ']') {
			number = 0;
			PopState(angle, start);
			pDC->MoveTo(start);
		}/* else if (at == '|') {
			int mx = Round(start.x + (m_iSegment*cos(_deg2rad(angle))));
			int my = Round(start.y + (m_iSegment*sin(_deg2rad(angle))));
			
			start.x = mx;
			start.y = my;

			pDC->LineTo(start);						
		} */else if (at >= '0' && at <= '9') {
			number = number * 10 + (at - '0');
		}
	}

	pDC->SelectObject(oldpen);
}

/////////////////////////////////////////////////////////////////////////////
// CLseView diagnostics

#ifdef _DEBUG
void CLseView::AssertValid() const
{
	CView::AssertValid();
}

void CLseView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CLseDoc* CLseView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CLseDoc)));
	return (CLseDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CLseView message handlers

void CLseView::OnInitialUpdate() 
{
	CView::OnInitialUpdate();
	
	CRect rect;
	GetClientRect(rect);
	m_cStartPoint = rect.CenterPoint();

	CWinApp *app = AfxGetApp();

	int r,g,b;
	r = app->GetProfileInt("Settings", "LineR", 255);
	g = app->GetProfileInt("Settings", "LineG", 255);
	b = app->GetProfileInt("Settings", "LineB", 255);

	m_crStart = RGB(r,g,b);

	r = app->GetProfileInt("Settings", "Line2R", 255);
	g = app->GetProfileInt("Settings", "Line2G", 255);
	b = app->GetProfileInt("Settings", "Line2B", 255);

	m_crEnd = RGB(r,g,b);

	r = app->GetProfileInt("Settings", "BackR", 0);
	g = app->GetProfileInt("Settings", "BackG", 0);
	b = app->GetProfileInt("Settings", "BackB", 0);

	m_crBackground = RGB(r,g,b);

	m_bMouseWheel = app->GetProfileInt("Settings", "WheelEnable", 0);
	m_iMouseDepth = app->GetProfileInt("Settings", "MouseDepth", 5);
	m_bPropOnLoad = app->GetProfileInt("Settings", "PropOnLoad", false);
}

void CLseView::OnEditLsystem() 
{
	CLSysDlg dlg;

	dlg.SetAxiomAndRules(&m_pszAxiom, &m_pszRules[0], &m_fAngle);
	dlg.SetOptions(&m_fInitAngle, &m_fStepSize, &m_iSegment);
	dlg.SetName(&m_pszName);

	if (dlg.DoModal() == IDOK) {
		reinterpret_cast<CLseDoc *>(GetDocument())->SetDescription(m_pszName);
		reinterpret_cast<CLseDoc *>(GetDocument())->SetTitle("");
		Invalidate();
	}
}

void CLseView::InitRulesArray() 
{
	for (int i=0; i<26; i++) { m_pszRules[i] = ""; }
}

void CLseView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	m_cDragPoint = point;
	m_bMouseDragging = true;

	SetCapture();
	
	CView::OnLButtonDown(nFlags, point);
}

void CLseView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	m_bMouseDragging = false;
	
	ReleaseCapture();

	CView::OnLButtonUp(nFlags, point);
}

void CLseView::OnMouseMove(UINT nFlags, CPoint point) 
{
	if (m_bMouseDragging) {
		if (nFlags & MK_CONTROL) {
			m_cStartPoint = point;
		} else {
			CPoint offset = point - m_cDragPoint;
			m_cStartPoint.Offset(offset);
			m_cDragPoint = point;
		}

		Invalidate();
	}

	CView::OnMouseMove(nFlags, point);
}

BOOL CLseView::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt) 
{
	if (!m_bMouseWheel) return CView::OnMouseWheel(nFlags, zDelta, pt);

	if (nFlags & MK_CONTROL) {
		if (zDelta > 0) {
			m_iMaxDepth = (m_iMaxDepth > 0) ? m_iMaxDepth - 1 : 0;
		} else {
			m_iMaxDepth = (m_iMaxDepth < m_iMouseDepth) ? m_iMaxDepth + 1 : m_iMaxDepth;
		}
	} else {
		if (zDelta > 0) {
			m_iSegment += 2;
		} else {
			m_iSegment -= 2;
		}

		if (m_iSegment % 2 != 0) m_iSegment++;
	}
	
	CalculateInfo();

	Invalidate();
	
	return CView::OnMouseWheel(nFlags, zDelta, pt);
}

#include "optionsdlg.h"
void CLseView::OnViewOptions() 
{
	COptionsDlg dlg;

	if (dlg.DoModal() == IDOK) {
		dlg.GetColours(m_crStart, m_crEnd, m_crBackground);

		CalculateInfo();

		CWinApp *app = AfxGetApp();

		app->WriteProfileInt("Settings", "LineR", GetRValue(m_crStart));
		app->WriteProfileInt("Settings", "LineG", GetGValue(m_crStart));
		app->WriteProfileInt("Settings", "LineB", GetBValue(m_crStart));
		
		app->WriteProfileInt("Settings", "Line2R", GetRValue(m_crEnd));
		app->WriteProfileInt("Settings", "Line2G", GetGValue(m_crEnd));
		app->WriteProfileInt("Settings", "Line2B", GetBValue(m_crEnd));

		app->WriteProfileInt("Settings", "BackR", GetRValue(m_crBackground));
		app->WriteProfileInt("Settings", "BackG", GetGValue(m_crBackground));
		app->WriteProfileInt("Settings", "BackB", GetBValue(m_crBackground));

		m_bMouseWheel = app->GetProfileInt("Settings", "WheelEnable", 0);
		m_iMouseDepth = app->GetProfileInt("Settings", "MouseDepth", 5);
		m_bPropOnLoad = app->GetProfileInt("Settings", "PropOnLoad", false);

		Invalidate();
	}
}

BOOL CLseView::OnEraseBkgnd(CDC* pDC) 
{
	return false;
	//	return CView::OnEraseBkgnd(pDC);
}

bool CLseView::LoadFile(CString file)
{
	ifstream data(file);

	int integer;
	float real;
	char rule;
	char line[256];

	InitRulesArray();
	m_fAngle = 90;
	m_fInitAngle = 0;
	m_fStepSize = 0.65f;
	m_iSegment = 6;
	m_iMaxDepth = 4;
	m_pszName = "";

	while (data.peek() != -1) {
		data.getline(line, 256, '=');

		if (stricmp(line,"axiom") == 0) {
			data.getline(line, 256);
			m_pszAxiom = line;
		} else if (stricmp(line, "segment") == 0) {
			data >> integer;
			m_iSegment = integer;
			data.getline(line, 1024);
		} else if (stricmp(line, "maxdepth") == 0) {
			data >> integer;
			m_iMaxDepth = integer;
			data.getline(line, 1024);
		} else if (stricmp(line, "stepsize") == 0) {
			data >> real;
			m_fStepSize = real;
			data.getline(line, 1024);
		} else if (stricmp(line, "angle") == 0) {
			data >> real;
			m_fAngle = real;
			data.getline(line, 1024);
		} else if (stricmp(line, "initangle") == 0) {
			data >> real;
			m_fInitAngle = real;
			data.getline(line, 1024);
		} else if (stricmp(line, "description") == 0) {
			data.getline(line, 1024);
			reinterpret_cast<CLseDoc *>(GetDocument())->SetDescription(line);
			m_pszName = line;
		} else {
			rule = line[0];
			if (rule >= 'A' && rule <= 'Z') {
				data.getline(line, 1024);
				m_pszRules[rule-'A'] = line;
			}
		}
	}

	data.close();

	CalculateInfo();

	if (m_bPropOnLoad) {
		CRect rect;
		GetClientRect(rect);
		m_cStartPoint = rect.CenterPoint();

		Invalidate();
		OnEditLsystem();
	}

	return true;
}
 
bool CLseView::SaveFile(CString file)
{
	ofstream data(file);

	data << "Description=" << m_pszName << endl;
	data << "angle=" << m_fAngle << endl;
	data << "initangle=" << m_fInitAngle << endl;
	data << "stepsize=" << m_fStepSize << endl;
	data << "segment=" << m_iSegment << endl;
	data << "maxdepth=" << m_iMaxDepth << endl;
	data << "axiom=" << m_pszAxiom << endl;
	
	for (int i=0; i<26; i++) {
		if (stricmp(m_pszRules[i], "") != 0) {
			data << char(i + 'A') << '=' << m_pszRules[i] << endl;
		}
	}

	data.close();

	return true;
}
 

void CLseView::OnEditDecreaseDepth() 
{
	m_iMaxDepth--;

	CalculateInfo();
	Invalidate();
}

void CLseView::OnEditIncreaseDepth() 
{
	m_iMaxDepth++;

	CalculateInfo();
	Invalidate();
}

void CLseView::OnEditCopy() 
{
	CClientDC dc(this);
	CDC bmpdc, memdc;
	CBitmap bmp, *oldbmp;
	CPen *oldpen, pen(PS_SOLID, 1, m_crLine);

	int cxDIB, cyDIB;								// The width/height of the bitmap
	
	CRect rect;
	GetClientRect(rect);
	cxDIB = rect.Width();
	cyDIB = rect.Height();

	bmpdc.CreateCompatibleDC(GetDC());
	memdc.CreateCompatibleDC(GetDC());
	bmp.CreateCompatibleBitmap(&dc, cxDIB, cyDIB);

	oldbmp = bmpdc.SelectObject(&bmp);
	oldpen = bmpdc.SelectObject(&pen);
	
//	CBrush br(RGB(0,0,0));
//	memdc.BitBlt(0,0,cxDIB,cyDIB,&bmpdc,0,0,SRCCOPY);
//	memdc.FrameRect(&CRect(0,0,cxDIB, cyDIB), &br);

	bmpdc.FillSolidRect(0,0,rect.Width(), rect.Height(), m_crBackground);
	bmpdc.MoveTo(m_cStartPoint);

	DrawLSystems(m_pszAxiom, &bmpdc, m_fInitAngle, 0);

	OpenClipboard();
	EmptyClipboard();
	SetClipboardData(CF_BITMAP, bmp.GetSafeHandle());
	CloseClipboard();

	bmp.Detach();
	bmpdc.SelectObject(oldbmp);
	bmpdc.SelectObject(oldpen);	
}

void CLseView::NewFile()
{
	InitRulesArray();

	m_fAngle = 90;
	m_fInitAngle = 0;
	m_fStepSize = 0.65f;
	m_iSegment = 6;
	m_iMaxDepth = 4;

	m_pszAxiom = "";
	m_pszName = "";
	
	for (int i=0; i<26; m_pszRules[i++] = "") ;
}

void CLseView::GenerateGradient(int iDepth)
{
	m_dr = (GetRValue(m_crEnd) - GetRValue(m_crStart)) / double(iDepth);
	m_dg = (GetGValue(m_crEnd) - GetGValue(m_crStart)) / double(iDepth);
	m_db = (GetBValue(m_crEnd) - GetBValue(m_crStart)) / double(iDepth);
}

void CLseView::CalculateInfo()
{
	CDC memdc;							// A temp DC.
	memdc.CreateCompatibleDC(GetDC());

	m_iSegments = 0;
	m_bCalculating = TRUE;

	DrawLSystems(m_pszAxiom, &memdc, m_fInitAngle, 0);
	
	if (m_iSegments == 0) {
		AfxMessageBox("Nothing will be drawn");
	}

	GenerateGradient(m_iSegments);

	m_bCalculating = FALSE;
}
